;;; SERIAL RECEIVE INTERRUPT ;;;;;;;;;;;;;;;;;;;;;;;;;;

irq_receive:
; Store context
    IN      safekeeper,SREG
;   IN      rxd_temp,SREG
;   PUSH    rxd_temp
; Check for break
    SBIC    UCR,RXB8
    RJMP    irq_receive_nodmxbreak
irq_receive_dmxbreak:
; Break detected;
; check if the current input start code valid flag is set,
; if so update the last input count
    SBRS    flags,current_input_startcode_valid
    RJMP    dmxbreak_update_pointer
; The configured input start code flag is set, so update last input count,
; calculated by subtracting the minimum input pointer index from the current input pointer index 
    SUBI    ptr_input_l,low(input_min)
; As the calculated number will be one too high, because the input pointer already advanced by one,
; the number will have to be decremented by one again in order to represent the actual input channel count
    DEC     ptr_input_l
    MOV     current_input_count,ptr_input_l
dmxbreak_update_pointer:
; Reinitialize DMX input pointer
    LDI     ptr_input_h,high(input_min)
    LDI     ptr_input_l,low(input_min)
; Clear UART data but don't store it
    IN      rxd_temp,UDR
    RJMP    irq_receive_end
irq_receive_nodmxbreak:
; Check input table limit
    CPI     ptr_input_h,high(input_max + 1)
    LDI     rxd_temp,low(input_max + 1)
    CPC     ptr_input_l,rxd_temp
    BRNE    irq_receive_receive_value
; Passed input table limit; clear UART data but don't store it
    IN      rxd_temp,UDR
    RJMP    irq_receive_end
irq_receive_receive_value:
; Check if this will be the first received byte in the packet (the start code)
    CPI     ptr_input_h,high(input_min)
    LDI     rxd_temp,low(input_min)
    CPC     ptr_input_l,rxd_temp
; Read value from UART
    IN      rxd_temp,UDR
    BRNE    check_startcode_flag
; Check if the input start code wildcard flag is set, if so also set current input startcode valid flag
    SBRC    flags,input_wildcard
    RJMP    irq_receive_valid_startcode
; The wildcard is disabled, so check the configured input start code against the current one
    CP      input_startcode,rxd_temp
    BREQ    irq_receive_valid_startcode
; The configured input start code didn't match with the current input start code,
; so clear the current input startcode valid flag (bit 2 in flags register)
    CLT
    RJMP    store_startcode_flag
irq_receive_valid_startcode:
; Set current input startcode valid flag (bit 2 in flags register)
    SET
store_startcode_flag:
; Actually update the current input startcode valid flag (bit 2 in flags register)
    BLD     flags,current_input_startcode_valid
check_startcode_flag:
; Check if the current input start code valid flag is set,
; if so store the received value in the DMX input table
    SBRC    flags,current_input_startcode_valid
    RJMP    valid_startcode_flag
; The configured input start code flag isn't set,
; so ignore this received value but increase the DMX input table pointer by one
    ADIW    ptr_input_l,1
    RJMP    irq_receive_end
valid_startcode_flag:
; Store value in input table
    ST      Y+,rxd_temp
; Reset timer 1 in order to be able to measure the time since the last DMX input break
; (dummy is still 0 here so that it could safely be used)
    OUT     TCNT1H,dummy
    OUT     TCNT1L,dummy
; Clear timer 1's overflow flag by writing a 1 to it
    LDI     rxd_temp,(1<<TOV1)|(0<<OCF1A)|(0<<OCF1B)|(0<<ICF1)|(0<<TOV0);!0b10000001
    OUT     TIFR,rxd_temp
irq_receive_end:
; Restore context
    OUT     SREG,safekeeper
;   POP     rxd_temp
;   OUT     SREG,rxd_temp
    RETI

;;; SERIAL TRANSMIT INTERRUPT ;;;;;;;;;;;;;;;;;;;;;;;;;

irq_transmit:
; Store context
    IN      safekeeper,SREG
;   IN      rxd_temp,SREG
;   PUSH    rxd_temp
; Check if a merge is in progress (bit in flags register)
    SBRC    flags,merge_in_progress
    RJMP    irq_transmit_merge_in_progress
; Check output counter; if passed do not send additional values
    CP      output_countdown,dummy
    BREQ    irq_transmit_after_limit
; Check output table limit; if passed do not send additional values
    CPI     ptr_output_h,high(output_max+1)
    LDI     txd_temp,low(output_max+1)
    CPC     ptr_output_l,txd_temp
    BRNE    irq_transmit_before_limit
irq_transmit_after_limit:
; Passed output counter or output table limit
; Check if merge flag (bit in flags register) is set,
    SBRC    flags,merge_status
    RJMP    irq_transmit_merge_flag
; The merge flag isn't set
irq_transmit_after_hard_limit:
; Reinitialize output pointer, generate BREAK & MAB and transmit first value (start code)
    LDI     ptr_output_h,high(output_min)
    LDI     ptr_output_l,low(output_min)
; Start transmitting of break
    RCALL   transmit_break_mab
; Break will be ended in timer interrupt handler, in which the MAB and first character will also be sent
    RJMP    irq_transmit_end
irq_transmit_merge_flag:
; The merge flag is set, so reinitialize output pointer to address of first input value (not the start code)
; (so in fact the output pointer temporarily becomes an input pointer)
    LDI     ptr_output_h,high(input_min + 1)
    LDI     ptr_output_l,low(input_min + 1)
; Set merge in progress flag (bit 5 in flags register)
    SET
    BLD     flags,merge_in_progress
irq_transmit_merge_in_progress:
; A merge is currently in progress, so check input table limit and if it was
; passed do not send additional values, otherwise output the next input value
    CPI     ptr_output_h,high(input_max+1)
    LDI     txd_temp,low(input_max+1)
    CPC     ptr_output_l,txd_temp
    BRNE    irq_transmit_before_limit
; The input table limit was passed; clear the merge in progress flag
; Clear merge in progress flag
    CLT
    BLD     flags,merge_in_progress
; Make sure that DMX output pointer is reinitialized and that a new break is transmitted
    RJMP    irq_transmit_after_hard_limit
irq_transmit_before_limit:
; Read value from output table
    LD      txd_temp,X+
; Write new value to UART
    OUT     UDR,txd_temp
; Decrease channel counter
    MOV     txd_temp,output_countdown
    DEC     txd_temp
    MOV     output_countdown,txd_temp
irq_transmit_end:
; Restore context
    OUT     SREG,safekeeper
;   POP     rxd_temp
;   OUT     SREG,rxd_temp
    RETI

;;; TIMER 0 INTERRUPT ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

irq_timer0:                         ; %3 cycles to RJMP to interrupt handler
; Store context
    IN      safekeeper,SREG
;   IN      rxd_temp,SREG
;   PUSH    rxd_temp
; Stop timer 1 by clearing CS01 bit of TCCR0
    IN      tmr_temp,TCCR0          ; %1 cycle
    CBR     tmr_temp,2              ; %1 cycle
    OUT     TCCR0,tmr_temp          ; %1 cycle
; Complement the 14 cycles (marked with % in irq_timer0 and transmit_break_mab)
; to prolong the BREAK with two additional microseconds
    NOP                             ; 1 cycle
    NOP                             ; 1 cycle
    set_txd                         ; %2 cycles
; Reenable transmitter
    SBI     UCR,TXEN                ; *2 cycles
transmit_mab:
    MOV     tmr_temp,mab_time       ; *1 cycle
; Delay for 1 microsec for the configured number of times
transmit_mab_delay:
    DEC     tmr_temp                ; 1 cycle
    NOP                             ; 1 cycle
    NOP                             ; 1 cycle
    NOP                             ; 1 cycle
    NOP                             ; 1 cycle
    NOP                             ; 1 cycle
    BRNE    transmit_mab_delay      ; 2 cycles (if there is a jump) or 1 cycle (if there's no jump)
; Complement the missing cycle caused by the BRNE taking only 1 cycle once
    NOP                             ; 1 cycle
; Complement the 6 cycles (marked with *) to prolong the MAB with an additional microsecond
    NOP                             ; 1 cycle
; Initialize output counter with output_count setting
    MOV     output_countdown,output_count   ; *1 cycle
; Read first value from output table (start code)
    LD      tmr_temp,X+             ; *3 cycles
; Write new value to UART, so that interrupts start to be triggered
    OUT     UDR,tmr_temp            ; 1 cycle
; Restore context
    OUT     SREG,safekeeper
;   POP     rxd_temp
;   OUT     SREG,rxd_temp
    RETI                            ; 4 cycles
